/*
 * File   : cdrom.c
 * Purpose: This file contains some routines to read information from
 *          a CD-ROM drive with an audio CD in it and routines to read
 *          a CDDB-Database.
 * Autor  : Andreas Theofilu
 *
 * Copyright (C) 1999 by Theos Soft.
 */

/* 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
/*
 * The following header file needs the library 'libcdaudio' to be
 * installed!
 */
#include <cdaudio.h>

#include <Xmt/Xmt.h>
#include <Xmt/Create.h>
#include <Xmt/Converters.h>
#include <Xmt/Procedures.h>
#include <Xmt/Layout.h>
#include <Xmt/Chooser.h>
#include <Xmt/InputField.h>
#include <Xmt/WidgetType.h>
#include <Xmt/Dialogs.h>

#include "platten.h"
#include "cdrom.h"

/*
 * Funktionen welche nur in diesem Modul verwendet werden.
 */
int CheckLibVersion ();
unsigned long cddbGetDiscID (struct disc_info *disc, char *device);

/*
 * Folgende Funktion liest die Informationen der angegebenen 'discid'
 * aus der CDDB-Datenbank. Damit auch die Spielzeiten der einzelnen
 * Tracks ermittelt werden koennen, mu die entsprechende CD eingelegt
 * sein. Die Routine ermittelt die DiscID direkt von der CD und
 * ergaenzt die fehlenden Daten.
 */
int cddbGetDiscInfo (cddb_title_t *cddb, unsigned long discid, char *device, char *path, char *category)
{
struct disc_info disc;
unsigned long cdid;
int fd, i;
char cdfile[256], hv0[256], *p;

	memset (cddb, 0, sizeof (cddb_title_t));
	cdid = cddbGetDiscID (&disc, device);

	if (cdid < 100L)
	   return (int)cdid;

	if (discid > 0L && cdid != discid)
	   return CD_EDISCID;

	/*
	 * Zusammensetzen des Pfades
	 */
	sprintf (&cdfile[0], "%s/%s/%08x", path, category, cdid);

	/*
	 * CDDB-Eintrag oeffnen
	 */
	if ((fd = open (cdfile, O_RDONLY, 0400)) == -1)
	   return CD_ECDDBEN;

	/*
	 * An dieser Stelle steht fest, das es sich um einen gueltigen
	 * CDDB-Eintrag handelt. Zuerst wird nun die DiscID zugewiesen,
	 * dann wird in einer Schleife die Datei ausgelesen und der
	 * Inhalt in die Struktur kopiert.
	 */
	cddb->discid = cdid;
	cddb->tracks = disc.disc_total_tracks;

	if (cddb->tracks > 30)
	   cddb->tracks = 30;

	while (ReadLine (fd, &hv0[0], sizeof (hv0)) > 0)
	{
	   if (hv0[0] == '#')
	      continue;

	   if ((p = strchr (hv0, ':')) == NULL) 
	      if ((p = strchr (hv0, '=')) == NULL)
	         continue;

	   *p = 0;
	   p++;

	   if (!*p)
	      continue;

	   if (!strncmp (hv0, "DTITLE", 6))
	      strncpy (cddb->title, ltrim (rtrim (p)), 80);

	   if (!strncmp (hv0, "TTITLE", 6))
	   {
	      i = atoi (&hv0[6]);

	      if (i >= 0 && i < cddb->tracks)
		 strncpy (cddb->entry[i].song, ltrim (rtrim (p)), 80);
	   }

	   if (!strncmp (hv0, "EXTT", 4))
	   {
	      i = atoi (&hv0[4]);

	      if (i >= 0 && i < cddb->tracks)
		 strncpy (cddb->entry[i].singer, ltrim (rtrim (p)), 80);
	   }
	}

	close (fd);

	/*
	 * Alle im CDDB-Eintrag vorhandenen Daten wurden kopiert. Nun
	 * kopieren wir noch die Spielzeiten und Frames in die Struktur.
	 */
	for (i = 0; i < cddb->tracks; i++)
	{
	   sprintf (cddb->entry[i].zeit, "%02d:%02d",
			disc.disc_track[i].track_length.minutes,
			disc.disc_track[i].track_length.seconds);
	   cddb->entry[i].frame = disc.disc_track[i].track_pos.frames;
	}

	return 0;
}

unsigned long cddbGetDiscID (struct disc_info *disc, char *device)
{
int fd;
int i, a, t, n;
unsigned long discid;

        t = n = 0;

	if (CheckLibVersion ())
	   return CD_LIBVERS;

        if ((fd = cd_init_device (device)) == -1)
	   return CD_EDEVICE;

        if (cd_stat (fd, disc) == -1)
        {
           cd_finish (fd);
           return CD_ESTATUS; 
        }

        cd_finish (fd);

        if (!disc->disc_present)
	   return CD_NOAUDIO;

	/*
	 * Ab hier beginnt die Berechnung der DiscID. Als Basis werden
         * die Gesamtanzahl Tracks verwendet. Dazu werden dann die Spiel-
         * minuten und Sekunden zusammen mit eine Pruefziffer summiert.
	 * Untenstehende Formel wurde aus 'xmcd' von Ti Kan extrahiert.
         */
        a = disc->disc_total_tracks;

        for (i = 0; i < a; i++)
           n += cddb_sum ((disc->disc_track[i].track_pos.minutes * 60) +
                disc->disc_track[i].track_pos.seconds);

        t = ((disc->disc_track[a].track_pos.minutes * 60) +
             disc->disc_track[a].track_pos.seconds) -
             ((disc->disc_track[0].track_pos.minutes * 60) +
             disc->disc_track[0].track_pos.seconds);

        discid = ((n % 0xff) << 24 | t << 8 | disc->disc_total_tracks);
        return discid;
}

/*
 * Pruefsummenberechnung
 */
int cddb_sum (int n)
{
int ret;

        for (ret = 0; n > 0; n /= 10)
                ret += n % 10;

        return (ret);
}

int CheckLibVersion ()
{
char cdversion[20];
int Major, Minor, Patch, i;
char *p;

        cd_version (&cdversion[0], sizeof (cdversion));

	p = strtok (&cdversion[0], ".");
	i = 0;

	while (p)
	{
	   switch (i)
	   {
	      case 0: Major = atoi (p); break;
	      case 1: Minor = atoi (p); break;
	      case 2: Patch = atoi (p); break;
	   }

	   i++;
	   p = strtok (NULL, ".");
	}

	if (Major > 1)
	   return 1;

	if (Major != 0 || Minor != 99)
	   fprintf (stderr, "Warning! PLATTEN was made for libcdaudio 0.99.1!\n                 Hope it works with %d.%d.%d too.\n", Major, Minor, Patch);
}

